@@ -16,6 +16,18 @@
 #include <Python.h>
 #include <numpy/arrayobject.h>
 
+/* It's complicated, but if I do not include the C++ math headers, GCC
+   will complain about conversions from 'double' to 'float', whenever 'isnan'
+   is called in a templated function (but not outside templates).
+
+   The '#include <cmath>' seems to cure the problem.
+*/
+//#include <cmath>
+#define fc_isnan(X) ((X)!=(X))
+// There is Py_IS_NAN but it is so much slower on my x86_64 system with GCC!
+
+#include <limits> // for std::numeric_limits<...>::infinity()
+
 #include "fastcluster.cpp"
 
 // backwards compatibility
@@ -29,30 +41,29 @@
 class linkage_output {
 private:
   t_float * Z;
-  t_index pos;
 
 public:
   linkage_output(t_float * const Z) {
     this->Z = Z;
-    pos = 0;
   }
 
-  void append(const t_index node1, const t_index node2, const t_float dist, const t_float size) {
+  void append(const t_index node1, const t_index node2, const t_float dist,
+              const t_float size) {
     if (node1<node2) {
-      Z[pos++] = static_cast<t_float>(node1);
-      Z[pos++] = static_cast<t_float>(node2);
+      *(Z++) = static_cast<t_float>(node1);
+      *(Z++) = static_cast<t_float>(node2);
     }
     else {
-      Z[pos++] = static_cast<t_float>(node2);
-      Z[pos++] = static_cast<t_float>(node1);
+      *(Z++) = static_cast<t_float>(node2);
+      *(Z++) = static_cast<t_float>(node1);
     }
-    Z[pos++] = dist;
-    Z[pos++] = size;
+    *(Z++) = dist;
+    *(Z++) = size;
   }
 };
 
 /*
-  Generate the Scipy-specific output format for a dendrogram from the
+  Generate the SciPy-specific output format for a dendrogram from the
   clustering output.
 
   The list of merging steps can be sorted or unsorted.
@@ -61,10 +72,10 @@ public:
 // one of the clusters.
 #define size_(r_) ( ((r_<N) ? 1 : Z_(r_-N,3)) )
 
-template <bool sorted>
+template <const bool sorted>
 static void generate_SciPy_dendrogram(t_float * const Z, cluster_result & Z2, const t_index N) {
   // The array "nodes" is a union-find data structure for the cluster
-  // identites (only needed for unsorted cluster_result input).
+  // identities (only needed for unsorted cluster_result input).
   union_find nodes;
   if (!sorted) {
     std::stable_sort(Z2[0], Z2[N-1]);
@@ -74,21 +85,21 @@ static void generate_SciPy_dendrogram(t_float * const Z, cluster_result & Z2, co
   linkage_output output(Z);
   t_index node1, node2;
 
-  for (t_index i=0; i<N-1; i++) {
+  for (node const * NN=Z2[0]; NN!=Z2[N-1]; ++NN) {
     // Get two data points whose clusters are merged in step i.
     if (sorted) {
-      node1 = Z2[i]->node1;
-      node2 = Z2[i]->node2;
+      node1 = NN->node1;
+      node2 = NN->node2;
     }
     else {
       // Find the cluster identifiers for these points.
-      node1 = nodes.Find(Z2[i]->node1);
-      node2 = nodes.Find(Z2[i]->node2);
+      node1 = nodes.Find(NN->node1);
+      node2 = nodes.Find(NN->node2);
       // Merge the nodes in the union-find data structure by making them
       // children of a new node.
       nodes.Union(node1, node2);
     }
-    output.append(node1, node2, Z2[i]->dist, size_(node1)+size_(node2));
+    output.append(node1, node2, NN->dist, size_(node1)+size_(node2));
   }
 }
 
@@ -105,12 +116,41 @@ static PyMethodDef _fastclusterWrapMethods[] = {
   {NULL, NULL, 0, NULL}     /* Sentinel - marks the end of this structure */
 };
 
-// Tell Python about these methods.
+
+/* Tell Python about these methods.
+
+   Python 2.x and 3.x differ in their C APIs for this part.
+*/
+#if PY_VERSION_HEX >= 0x03000000
+
+static struct PyModuleDef fastclustermodule = {
+  PyModuleDef_HEAD_INIT,
+  "_fastcluster",
+  NULL, // no module documentation
+  -1,  /* size of per-interpreter state of the module,
+          or -1 if the module keeps state in global variables. */
+  _fastclusterWrapMethods
+};
+
+PyMODINIT_FUNC PyInit__fastcluster(void) {
+  PyObject * m;
+  m = PyModule_Create(&fastclustermodule);
+  if (!m) {
+    return NULL;
+  }
+  import_array();  // Must be present for NumPy. Called first after above line.
+  return m;
+}
+
+# else // Python 2.x
+
 PyMODINIT_FUNC init_fastcluster(void)  {
   (void) Py_InitModule("_fastcluster", _fastclusterWrapMethods);
   import_array();  // Must be present for NumPy. Called first after above line.
 }
 
+#endif // PY_VERSION
+
 static void end_allow_threads(PyThreadState * PythonThreadSave) {
   if (PythonThreadSave) { // Only restore if the state has been saved
       PyEval_RestoreThread(PythonThreadSave);
@@ -124,40 +164,47 @@ static void end_allow_threads(PyThreadState * PythonThreadSave) {
 
 static PyObject *linkage_wrap(PyObject * const self, PyObject * const args) {
   PyArrayObject * D, * Z;
-  long int N = 0;
+  long int N_ = 0;
   unsigned char method;
   PyThreadState * PythonThreadSave = NULL;
 
   try{
     // Parse the input arguments
     if (!PyArg_ParseTuple(args, "lO!O!b",
-                          &N,                // signed long integer
+                          &N_,                // signed long integer
                           &PyArray_Type, &D, // NumPy array
                           &PyArray_Type, &Z, // NumPy array
                           &method)) {        // unsigned char
       return NULL; // Error if the arguments have the wrong type.
     }
-    if (N < 1 ) {
+    if (N_ < 1 ) {
       // N must be at least 1.
       PyErr_SetString(PyExc_ValueError,
                       "At least one element is needed for clustering.");
       return NULL;
     }
 
-    // (1)
-    // The biggest index used below is 4*(N-2)+3, as an index to Z. This must fit
-    // into the data type used for indices.
-    // (2)
-    // The largest representable integer, without loss of precision, by a floating
-    // point number of type t_float is 2^T_FLOAT_MANT_DIG. Here, we make sure that
-    // all cluster labels from 0 to 2N-2 in the output can be accurately represented
-    // by a floating point number.
-    if (N > MAX_INDEX/4 ||
-        (N-1)>>(T_FLOAT_MANT_DIG-1) > 0) {
+    /*
+      (1)
+      The biggest index used below is 4*(N-2)+3, as an index to Z. This must
+      fit into the data type used for indices.
+      (2)
+      The largest representable integer, without loss of precision, by a
+      floating point number of type t_float is 2^T_FLOAT_MANT_DIG. Here, we
+      make sure that all cluster labels from 0 to 2N-2 in the output can be
+      accurately represented by a floating point number.
+
+      Conversion of N to 64 bits below is not really necessary but it prevents
+      a warning ("shift count >= width of type") on systems where "long int"
+      is 32 bits wide.
+    */
+    if (N_ > MAX_INDEX/4 ||
+        static_cast<int64_t>(N_-1)>>(T_FLOAT_MANT_DIG-1) > 0) {
       PyErr_SetString(PyExc_ValueError,
                       "Data is too big, index overflow.");
       return NULL;
     }
+    t_index N = static_cast<t_index>(N_);
 
     if (method>METHOD_METR_MEDIAN) {
       PyErr_SetString(PyExc_IndexError, "Invalid method index.");
@@ -181,8 +228,9 @@ static PyObject *linkage_wrap(PyObject * const self, PyObject * const args) {
     if (method==METHOD_METR_WARD ||
         method==METHOD_METR_CENTROID ||
         method==METHOD_METR_MEDIAN) {
-      for (std::ptrdiff_t i=0; i < static_cast<std::ptrdiff_t>(N)*(N-1)/2; i++)
-        D_[i] *= D_[i];
+      for (t_float * DD = D_; DD!=D_+static_cast<std::ptrdiff_t>(N)*(N-1)/2;
+           ++DD)
+        *DD *= *DD;
     }
 
     switch (method) {
@@ -223,6 +271,17 @@ static PyObject *linkage_wrap(PyObject * const self, PyObject * const args) {
       generate_SciPy_dendrogram<false>(Z_, Z2, N);
     }
 
+    /* Hack: prevent PyEval_RestoreThread from being called a second time
+       in the "catch" block, in case the first PyEval_RestoreThread in
+       end_allow_threads throws an exception.
+
+       (Can PyEval_RestoreThread throw an exception? I don't know, so I do this
+       to be on the safe side.)
+    */
+    PyThreadState * const PythonThreadSave2 = PythonThreadSave;
+    PythonThreadSave = NULL;
+    end_allow_threads(PythonThreadSave2);
+    Py_RETURN_NONE;
   } // try
   catch (const std::bad_alloc&) {
     end_allow_threads(PythonThreadSave);
@@ -230,17 +289,27 @@ static PyObject *linkage_wrap(PyObject * const self, PyObject * const args) {
   }
   catch(const std::exception& e){
     end_allow_threads(PythonThreadSave);
-    PyErr_SetString(PyExc_StandardError, e.what());
+    PyErr_SetString(PyExc_EnvironmentError, e.what());
     return NULL;
   }
+  catch(const nan_error&){
+    end_allow_threads(PythonThreadSave);
+    PyErr_SetString(PyExc_FloatingPointError, "NaN dissimilarity value.");
+    return NULL;
+  }
+  #ifdef FE_INVALID
+  catch(const fenv_error&){
+    end_allow_threads(PythonThreadSave);
+    PyErr_SetString(PyExc_FloatingPointError, "NaN dissimilarity value in intermediate results.");
+    return NULL;
+  }
+  #endif
   catch(...){
     end_allow_threads(PythonThreadSave);
-    PyErr_SetString(PyExc_StandardError,
+    PyErr_SetString(PyExc_EnvironmentError,
                     "C++ exception (unknown reason). Please send a bug report.");
     return NULL;
   }
-  PyEval_RestoreThread(PythonThreadSave);
-  Py_RETURN_NONE;
 }
 
 /*
@@ -270,8 +339,8 @@ enum {
   METRIC_KULSINSKI       = 18,
   METRIC_USER            = 19,
   METRIC_INVALID         = 20, // sentinel
-  METRIC_JACCARD_BOOL    = 21, // separate function for Jaccard metric on Boolean
-};                             // input data
+  METRIC_JACCARD_BOOL    = 21, // separate function for Jaccard metric on
+};                             // Boolean input data
 
 /*
    Helper class: Throw this if calling the Python interpreter from within
@@ -315,7 +384,7 @@ public:
                         bool temp_point_array)
     : Xa(reinterpret_cast<t_float *>(PyArray_DATA(Xarg))),
       dim(PyArray_DIM(Xarg, 1)),
-      N(PyArray_DIM(Xarg, 0)),
+      N(static_cast<t_index>(PyArray_DIM(Xarg, 0))),
       members(members),
       postprocessfn(NULL),
       V(NULL)
@@ -330,14 +399,14 @@ public:
       case METRIC_SEUCLIDEAN:
         if (extraarg==NULL) {
           PyErr_SetString(PyExc_TypeError,
-                          "The 'seuclidean' metric needs a variance parameter.");
+              "The 'seuclidean' metric needs a variance parameter.");
           throw pythonerror();
         }
-        V  = reinterpret_cast<PyArrayObject *>(PyArray_FromAny(extraarg,
-                                               PyArray_DescrFromType(NPY_DOUBLE),
-                                               1, 1,
-                                               NPY_ARRAY_CARRAY_RO,
-                                               NULL));
+        V = reinterpret_cast<PyArrayObject *>(PyArray_FromAny(extraarg,
+                PyArray_DescrFromType(NPY_DOUBLE),
+                1, 1,
+                NPY_ARRAY_CARRAY_RO,
+                NULL));
         if (PyErr_Occurred()) {
           throw pythonerror();
         }
@@ -351,7 +420,7 @@ public:
         postprocessfn = &cluster_result::sqrt;
         break;
       case METRIC_SQEUCLIDEAN:
-        distfn = &python_dissimilarity::sqeuclidean;
+        distfn = &python_dissimilarity::sqeuclidean<false>;
         break;
       case METRIC_CITYBLOCK:
         set_cityblock();
@@ -367,9 +436,9 @@ public:
         postprocessfn = &cluster_result::plusone;
         // precompute norms
         precomputed.init(N);
-        for (t_index i=0; i<N; i++) {
+        for (t_index i=0; i<N; ++i) {
           t_float sum=0;
-          for (t_index k=0; k<dim; k++) {
+          for (t_index k=0; k<dim; ++k) {
             sum += X(i,k)*X(i,k);
           }
           precomputed[i] = 1/sqrt(sum);
@@ -438,9 +507,9 @@ public:
         distfn = &python_dissimilarity::kulsinski;
         postprocessfn = &cluster_result::plusone;
         precomputed.init(N);
-        for (t_index i=0; i<N; i++) {
+        for (t_index i=0; i<N; ++i) {
           t_index sum=0;
-          for (t_index k=0; k<dim; k++) {
+          for (t_index k=0; k<dim; ++k) {
             sum += Xb(i,k);
           }
           precomputed[i] = -.5/static_cast<t_float>(sum);
@@ -492,7 +561,7 @@ public:
   void merge(const t_index i, const t_index j, const t_index newnode) const {
     t_float const * const Pi = i<N ? Xa+i*dim : Xnew+(i-N)*dim;
     t_float const * const Pj = j<N ? Xa+j*dim : Xnew+(j-N)*dim;
-    for(t_index k=0; k<dim; k++) {
+    for(t_index k=0; k<dim; ++k) {
       Xnew[(newnode-N)*dim+k] = (Pi[k]*static_cast<t_float>(members[i]) +
                                  Pj[k]*static_cast<t_float>(members[j])) /
         static_cast<t_float>(members[i]+members[j]);
@@ -500,10 +569,11 @@ public:
     members[newnode] = members[i]+members[j];
   }
 
-  void merge_weighted(const t_index i, const t_index j, const t_index newnode) const {
+  void merge_weighted(const t_index i, const t_index j, const t_index newnode)
+    const {
     t_float const * const Pi = i<N ? Xa+i*dim : Xnew+(i-N)*dim;
     t_float const * const Pj = j<N ? Xa+j*dim : Xnew+(j-N)*dim;
-    for(t_index k=0; k<dim; k++) {
+    for(t_index k=0; k<dim; ++k) {
       Xnew[(newnode-N)*dim+k] = (Pi[k]+Pj[k])*.5;
     }
   }
@@ -511,7 +581,7 @@ public:
   void merge_inplace(const t_index i, const t_index j) const {
     t_float const * const Pi = Xa+i*dim;
     t_float * const Pj = Xa+j*dim;
-    for(t_index k=0; k<dim; k++) {
+    for(t_index k=0; k<dim; ++k) {
       Pj[k] = (Pi[k]*static_cast<t_float>(members[i]) +
                Pj[k]*static_cast<t_float>(members[j])) /
         static_cast<t_float>(members[i]+members[j]);
@@ -522,7 +592,7 @@ public:
   void merge_inplace_weighted(const t_index i, const t_index j) const {
     t_float const * const Pi = Xa+i*dim;
     t_float * const Pj = Xa+j*dim;
-    for(t_index k=0; k<dim; k++) {
+    for(t_index k=0; k<dim; ++k) {
       Pj[k] = (Pi[k]+Pj[k])*.5;
     }
   }
@@ -536,15 +606,16 @@ public:
   inline t_float ward(const t_index i, const t_index j) const {
     t_float mi = static_cast<t_float>(members[i]);
     t_float mj = static_cast<t_float>(members[j]);
-    return sqeuclidean(i,j)*mi*mj/(mi+mj);
+    return sqeuclidean<true>(i,j)*mi*mj/(mi+mj);
   }
 
   inline t_float ward_initial(const t_index i, const t_index j) const {
     // alias for sqeuclidean
     // Factor 2!!!
-    return sqeuclidean(i,j);
+    return sqeuclidean<true>(i,j);
   }
 
+  // This method must not produce NaN if the input is non-NaN.
   inline static t_float ward_initial_conversion(const t_float min) {
     return min*.5;
   }
@@ -555,21 +626,30 @@ public:
     return sqeuclidean_extended(i,j)*mi*mj/(mi+mj);
   }
 
+  /* We need two variants of the Euclidean metric: one that does not check
+     for a NaN result, which is used for the initial distances, and one which
+     does, for the updated distances during the clustering procedure.
+  */
+  template <const bool check_NaN>
   t_float sqeuclidean(const t_index i, const t_index j) const {
     t_float sum = 0;
     /*
-    for (t_index k=0; k<dim; k++) {
-        t_float diff = X(i,k) - X(j,k);
-        sum += diff*diff;
-    }
+      for (t_index k=0; k<dim; ++k) {
+      t_float diff = X(i,k) - X(j,k);
+      sum += diff*diff;
+      }
     */
     // faster
     t_float const * Pi = Xa+i*dim;
     t_float const * Pj = Xa+j*dim;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       t_float diff = Pi[k] - Pj[k];
       sum += diff*diff;
     }
+    if (check_NaN) {
+      if (fc_isnan(sum))
+        throw(nan_error());
+    }
     return sum;
   }
 
@@ -577,10 +657,12 @@ public:
     t_float sum = 0;
     t_float const * Pi = i<N ? Xa+i*dim : Xnew+(i-N)*dim; // TBD
     t_float const * Pj = j<N ? Xa+j*dim : Xnew+(j-N)*dim;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       t_float diff = Pi[k] - Pj[k];
       sum += diff*diff;
     }
+    if (fc_isnan(sum))
+      throw(nan_error());
     return sum;
   }
 
@@ -612,7 +694,7 @@ private:
   }
 
   void set_euclidean() {
-    distfn = &python_dissimilarity::sqeuclidean;
+    distfn = &python_dissimilarity::sqeuclidean<false>;
     postprocessfn = &cluster_result::sqrt;
   }
 
@@ -626,7 +708,7 @@ private:
 
   t_float seuclidean(const t_index i, const t_index j) const {
     t_float sum = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       t_float diff = X(i,k)-X(j,k);
       sum += diff*diff/V_data[k];
     }
@@ -635,7 +717,7 @@ private:
 
   t_float cityblock(const t_index i, const t_index j) const {
     t_float sum = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       sum += fabs(X(i,k)-X(j,k));
     }
     return sum;
@@ -643,7 +725,7 @@ private:
 
   t_float minkowski(const t_index i, const t_index j) const {
     t_float sum = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       sum += pow(fabs(X(i,k)-X(j,k)),postprocessarg);
     }
     return sum;
@@ -651,7 +733,7 @@ private:
 
   t_float chebychev(const t_index i, const t_index j) const {
     t_float max = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       t_float diff = fabs(X(i,k)-X(j,k));
       if (diff>max) {
         max = diff;
@@ -662,7 +744,7 @@ private:
 
   t_float cosine(const t_index i, const t_index j) const {
     t_float sum = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       sum -= X(i,k)*X(j,k);
     }
     return sum*precomputed[i]*precomputed[j];
@@ -670,7 +752,7 @@ private:
 
   t_float hamming(const t_index i, const t_index j) const {
     t_float sum = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       sum += (X(i,k)!=X(j,k));
     }
     return sum;
@@ -681,7 +763,7 @@ private:
   t_float jaccard(const t_index i, const t_index j) const {
     t_index sum1 = 0;
     t_index sum2 = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       sum1 += (X(i,k)!=X(j,k));
       sum2 += ((X(i,k)!=0) || (X(j,k)!=0));
     }
@@ -690,7 +772,7 @@ private:
 
   t_float canberra(const t_index i, const t_index j) const {
     t_float sum = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       t_float numerator = fabs(X(i,k)-X(j,k));
       sum += numerator==0 ? 0 : numerator / (fabs(X(i,k)) + fabs(X(j,k)));
     }
@@ -717,7 +799,7 @@ private:
   t_float braycurtis(const t_index i, const t_index j) const {
     t_float sum1 = 0;
     t_float sum2 = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       sum1 += fabs(X(i,k)-X(j,k));
       sum2 += fabs(X(i,k)+X(j,k));
     }
@@ -727,7 +809,7 @@ private:
   t_float mahalanobis(const t_index i, const t_index j) const {
     // V_data contains the product X*VI
     t_float sum = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       sum += (V_data[i*dim+k]-V_data[j*dim+k])*(X(i,k)-X(j,k));
     }
     return sum;
@@ -742,7 +824,7 @@ private:
   void nbool_correspond(const t_index i, const t_index j) const {
     NTT = 0;
     NXO = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       NTT += (Xb(i,k) &  Xb(j,k)) ;
       NXO += (Xb(i,k) ^  Xb(j,k)) ;
     }
@@ -752,26 +834,26 @@ private:
     NTT = 0;
     NXO = 0;
     NTF = 0;
-    for (t_index k=0; k<dim; k++) {
+    for (t_index k=0; k<dim; ++k) {
       NTT += (Xb(i,k) &  Xb(j,k)) ;
       NXO += (Xb(i,k) ^  Xb(j,k)) ;
       NTF += (Xb(i,k) & ~Xb(j,k)) ;
     }
     NTF *= (NXO-NTF); // NTFFT
-    NTT *= (dim-NTT-NXO); // NFFTT
+    NTT *= (static_cast<t_index>(dim)-NTT-NXO); // NFFTT
   }
 
   void nbool_correspond_xo(const t_index i, const t_index j) const {
     NXO = 0;
-    for (t_index k=0; k<dim; k++) {
-      NXO += (Xb(i,k) ^  Xb(j,k)) ;
+    for (t_index k=0; k<dim; ++k) {
+      NXO += (Xb(i,k) ^ Xb(j,k)) ;
     }
   }
 
   void nbool_correspond_tt(const t_index i, const t_index j) const {
     NTT = 0;
-    for (t_index k=0; k<dim; k++) {
-      NTT += (Xb(i,k) &  Xb(j,k)) ;
+    for (t_index k=0; k<dim; ++k) {
+      NTT += (Xb(i,k) & Xb(j,k)) ;
     }
   }
 
@@ -845,8 +927,8 @@ static PyObject *linkage_vector_wrap(PyObject * const self, PyObject * const arg
       PyErr_SetString(PyExc_ValueError,
                       "The input array must be two-dimensional.");
     }
-    npy_intp const N = PyArray_DIM(X, 0);
-    if (N < 1 ) {
+    npy_intp const N_ = PyArray_DIM(X, 0);
+    if (N_ < 1 ) {
       // N must be at least 1.
       PyErr_SetString(PyExc_ValueError,
                       "At least one element is needed for clustering.");
@@ -860,20 +942,27 @@ static PyObject *linkage_vector_wrap(PyObject * const self, PyObject * const arg
       return NULL;
     }
 
-    // (1)
-    // The biggest index used below is 4*(N-2)+3, as an index to Z. This must fit
-    // into the data type used for indices.
-    // (2)
-    // The largest representable integer, without loss of precision, by a floating
-    // point number of type t_float is 2^T_FLOAT_MANT_DIG. Here, we make sure that
-    // all cluster labels from 0 to 2N-2 in the output can be accurately represented
-    // by a floating point number.
-    if (N > MAX_INDEX/4 ||
-        (N-1)>>(T_FLOAT_MANT_DIG-1) > 0) {
+    /*
+      (1)
+      The biggest index used below is 4*(N-2)+3, as an index to Z. This must
+      fit into the data type used for indices.
+      (2)
+      The largest representable integer, without loss of precision, by a
+      floating point number of type t_float is 2^T_FLOAT_MANT_DIG. Here, we
+      make sure that all cluster labels from 0 to 2N-2 in the output can be
+      accurately represented by a floating point number.
+
+      Conversion of N to 64 bits below is not really necessary but it prevents
+      a warning ("shift count >= width of type") on systems where "int" is 32
+      bits wide.
+    */
+    if (N_ > MAX_INDEX/4 || dim > MAX_INDEX ||
+        static_cast<int64_t>(N_-1)>>(T_FLOAT_MANT_DIG-1) > 0) {
       PyErr_SetString(PyExc_ValueError,
                       "Data is too big, index overflow.");
       return NULL;
     }
+    t_index N = static_cast<t_index>(N_);
 
     cluster_result Z2(N-1);
 
@@ -967,6 +1056,7 @@ static PyObject *linkage_vector_wrap(PyObject * const self, PyObject * const arg
     PyThreadState * const PythonThreadSave2 = PythonThreadSave;
     PythonThreadSave = NULL;
     end_allow_threads(PythonThreadSave2);
+    Py_RETURN_NONE;
   } // try
   catch (const std::bad_alloc&) {
     end_allow_threads(PythonThreadSave);
@@ -974,7 +1064,12 @@ static PyObject *linkage_vector_wrap(PyObject * const self, PyObject * const arg
   }
   catch(const std::exception& e){
     end_allow_threads(PythonThreadSave);
-    PyErr_SetString(PyExc_StandardError, e.what());
+    PyErr_SetString(PyExc_EnvironmentError, e.what());
+    return NULL;
+  }
+  catch(const nan_error&){
+    end_allow_threads(PythonThreadSave);
+    PyErr_SetString(PyExc_FloatingPointError, "NaN dissimilarity value.");
     return NULL;
   }
   catch(const pythonerror){
@@ -982,9 +1077,8 @@ static PyObject *linkage_vector_wrap(PyObject * const self, PyObject * const arg
   }
   catch(...){
     end_allow_threads(PythonThreadSave);
-    PyErr_SetString(PyExc_StandardError,
+    PyErr_SetString(PyExc_EnvironmentError,
                     "C++ exception (unknown reason). Please send a bug report.");
     return NULL;
   }
-  Py_RETURN_NONE;
 }
